/* sdsl - succinct data structures library
    Copyright (C) 2008 Simon Gog 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file csa_sada_prac.hpp
    \brief csa_sada_prac.hpp contains an implementation of the compressed suffix array. 
	\author Simon Gog
*/
#ifndef INCLUDED_SDSL_CSA_SADA_PRAC
#define INCLUDED_SDSL_CSA_SADA_PRAC

#include "enc_vector_prac.hpp" 
#include "int_vector.hpp"
#include "algorithms.hpp"
#include "iterators.hpp"
#include "util.hpp"
#include <iostream>
#include <algorithm>
#include <cassert>
#include <cstring> // for strlen
#include <iomanip>
#include <iterator>

namespace sdsl{

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens,  uint8_t fixedIntWidth>
class csa_sada_prac;

template<uint8_t fixedIntWidth>
struct csa_sada_prac_trait{
	typedef int_vector<0> int_vector_type;
};

template<>
struct csa_sada_prac_trait<32>{
	typedef int_vector<32> int_vector_type;
};

template<>
struct csa_sada_prac_trait<64>{
	typedef int_vector<64> int_vector_type;
};


//! A class for the Compressed Suffix Array (CSA) proposed by Sadakane for practical implementation.
/*! The CSA is parameterized with an EncVector and the sample density SampleDens (\f$s_{SA}\f$).
  * I.e. every \f$s_{SA}th\f$ value from the original suffix array is explicitely stored with \f$\log n\f$ bits.
  * 
  * The EncVector (default is sdsl::enc_vector_prac) holds the \f$\Psi\f$-function an can be paremetrized with \f$s_{\Psi}\f$.
  * \todo example, code example
  *  \sa csa_sada_theo
  * @ingroup csa
 */
template<class EncVector = enc_vector_prac<>, uint32_t SampleDens = 12, uint32_t InvSampleDens = 12,  uint8_t fixedIntWidth = 32>
class csa_sada_prac{
	public:
	typedef uint64_t											 value_type;	// STL Container requirement
	typedef random_access_const_iterator<csa_sada_prac> const_iterator;// STL Container requirement
	typedef const_iterator 										 iterator;		// STL Container requirement
	typedef const value_type									 const_reference;
	typedef const_reference										 reference;
	typedef const_reference*									 pointer;
	typedef const pointer										 const_pointer;
	typedef size_t 												 size_type;		// STL Container requirement
	typedef ptrdiff_t  											 difference_type; // STL Container requirement
	typedef EncVector											 enc_vector_type;
	typedef const unsigned char*								 pattern_type;
//	static const uint32_t sample_dens = SampleDens;
	private:
		EncVector m_psi;  // psi function
		typename csa_sada_prac_trait<fixedIntWidth>::int_vector_type m_sa_sample; // suffix array samples
		typename csa_sada_prac_trait<fixedIntWidth>::int_vector_type m_isa_sample; // inverse suffix array samples
		unsigned char	m_char2comp[256];
		unsigned char 	m_comp2char[256];
		size_type       m_C[257];
		uint8_t			m_sigma;
		uint32_t m_sample_dens; // additional to SampleDens value

		template<typename RandomAccessContainer>
		void construct(const RandomAccessContainer &sa, const unsigned char *str);

		template<typename RandomAccessContainer>
		void construct(RandomAccessContainer &sa, const unsigned char *str);

		template<typename RandomAccessContainer>
		void construct_samples(const RandomAccessContainer &sa, const unsigned char *str);

		void setText(const unsigned char *str, size_type len);
	public:
		const unsigned char *char2comp;
		const unsigned char *comp2char;
		const size_type *C;
		const uint8_t &sigma;
		const EncVector &psi;
		//! Default Constructor
		csa_sada_prac():char2comp(m_char2comp), comp2char(m_comp2char), C(m_C), sigma(m_sigma) ,psi(m_psi){}
		//! Default Destructor
		~csa_sada_prac(){}
		//! Copy constructor
		csa_sada_prac(const csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth> &csa):m_psi(csa.m_psi), m_sa_sample(csa.m_sa_sample), m_isa_sample(csa.m_isa_sample),  char2comp(m_char2comp), comp2char(m_comp2char), C(m_C), sigma(m_sigma), psi(m_psi)
		{ 
			m_sigma			= csa.m_sigma;
			for(int i=0; i<256; ++i){
				m_char2comp[i] 	= csa.m_char2comp[i];
				m_comp2char[i]	= csa.m_comp2char[i];
				m_C[i]			= csa.m_C[i];
			}
			m_C[256] = csa.m_C[256];
		}

		//! Construct the csa_sada_prac from another (compressed) suffix array and the original text
		/*! \param sa  The existing (compressed) suffix array for the string str.
		 *	\param str The text for which sa was build. 
		 */
		template<typename RandomAccessContainer>
		csa_sada_prac(const RandomAccessContainer &sa, const unsigned char *str);

		//! Construct the csa_sada_prac from another (compressed) suffix array and the original text
		/*! \param sa  The existing (compressed) suffix array for the string str.
		 *	\param str The text for which the suffix array sa was build. 
		 */
		template<typename RandomAccessContainer>
		csa_sada_prac(RandomAccessContainer &sa, const unsigned char *str);


		//! Constructor for the CSA taking a string for that the CSA should be calculated
		/*!	\param str The text for which the CSA should be constructed.
		 */
		csa_sada_prac(const unsigned char *str);

		//! Number of elements in the \f$\CSA\f$.
		/*! Required for the Container Concept of the STL.
		 *  \sa max_size, empty
		 *  \par Time complexity
		 *      \f$ \Order{1} \f$
		 */
		size_type size()const{ return m_psi.size();}

		//! Returns the largest size that csa_sada_prac can ever have.
		/*! Required for the Container Concept of the STL.
		 *  \sa size
		 */
		static size_type max_size(){ return EncVector::max_size(); }

		//! Returns if the data strucutre is empty.
		/*! Required for the Container Concept of the STL.A
		 * \sa size
		 */
		bool empty()const{ return m_psi.empty();}

		//! Swap method for csa_sada_prac
		/*! The swap method can be defined in terms of assignment.
			This requires three assignments, each of which, for a container type, is linear
			in the container's size. In a sense, then, a.swap(b) is redundant.
			This implementation guaranties a run-time complexity that is constant rather than linear.
			\param csa csa_sada_prac to swap.

			Required for the Assignable Conecpt of the STL.
		  */
		void swap(csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth> &csa);

		//! Returns a const_iterator to the first element.
		/*! Required for the STL Container Concept.
		 *  \sa end
		 */
		const_iterator begin()const;

		//! Returns a const_iterator to the element after the last element.
		/*! Required for the STL Container Concept.
		 *  \sa begin.
		 */
		const_iterator end()const;

		//! []-operator
		/*! \param i Index of the value. \f$ i \in [0..size()-1]\f$.
		 * Required for the STL Random Access Container Concept.
		 * \par Time complexity
		 *      \f$ \Order{s_{SA}\cdot t_{\Psi}} \f$, where every \f$s_{SA}\f$th suffix array entry is sampled and \f$t_{\Psi}\f$ 
		 *           is the access time for an element in the \f$\Psi\f$-function.
		 */
		inline value_type operator[](size_type i)const;

		//! ()-operator return inverse suffix array values
		/*! \param i Index of the value. \f$ i \in [0..size()-1]\f$.
		 *   \par Time complexity
		 *      \f$ \Order{s_{SA^{-1}}\cdot t_{\Psi}} \f$, where every \f$s_{SA^{-1}}\f$th suffix array entry is sampled and \f$t_{\Psi}\f$ 
		 *           is the access time for an element in the \f$\Psi\f$-function.
		 */
		inline value_type operator()(size_type i)const;

		//! Assignment Operator.
		/*!
		 *	Required for the Assignable Concept of the STL.
		 */
		csa_sada_prac& operator=(const csa_sada_prac &csa);

		//! Equality Operator
		/*! Two Instances of csa_sada_prac are equal if
		 *  all their members are equal.
		 *  \par Required for the Equality Comparable Concept of the STL.
		 *  \sa operator!=
		 */
		bool operator==(const csa_sada_prac &csa)const;

		//! Unequality Operator
		/*! Two Instances of csa_sada_prac are equal if
		 *  not all their members are equal.
		 *  \par Required for the Equality Comparable Concept of the STL.
		 *  \sa operator==
		 */
		bool operator!=(const csa_sada_prac &csa)const;

		//! Serialize to a stream.
		/*! \param out Outstream to write the data structure.
		 *  \return The number of written bytes.
		 */
		size_type serialize(std::ostream &out) const;

		//! Load from a stream.
		/*! \param in Inputstream to load the data structure from.
		 */
		void load(std::istream &in);

		uint32_t get_sample_dens() const;
		void set_sample_dens(const uint32_t sample_dens);

		uint32_t get_psi_sample_dens() const;
		void set_psi_sample_dens(const uint32_t sample_dens);
};

// == template functions ==

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::csa_sada_prac(const unsigned char *str):char2comp(m_char2comp), comp2char(m_comp2char), C(m_C), sigma(m_sigma), psi(m_psi){
	size_type n = strlen((const char*)str);
	int_vector<> sa(n+1, 0, bit_magic::l1BP(n+1)+1);
	algorithm::calculate_sa(str, n+1, sa);	 // calculate the suffix array sa of str
	assert(sa.size() == n+1);
	setText(str, n+1);
	construct(sa, str);
	if( n+1 > 0 and n+1 != size() )
		throw std::logic_error("csa_sada_prac: text size differ with sa size!");
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
template<typename RandomAccessContainer>
csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::csa_sada_prac(const RandomAccessContainer &sa, const unsigned char *str):char2comp(m_char2comp), comp2char(m_comp2char),C(m_C), sigma(m_sigma), psi(m_psi){
	size_type n = 1;
	if(str != NULL){
		n = strlen((const char*)str);
	}
	setText(str, n+1);
	assert(sa.size() == n+1);
	construct(sa, str);
	if( n+1 > 0 and n+1 != size() )
		throw std::logic_error(util::demangle(typeid(this).name())+": text size differ with sa size!");
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
template<typename RandomAccessContainer>
csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::csa_sada_prac(RandomAccessContainer &sa, const unsigned char *str):char2comp(m_char2comp), comp2char(m_comp2char),C(m_C), sigma(m_sigma), psi(m_psi){
	size_type n = 1;
	if(str != NULL){
		n = strlen((const char*)str);
	}
	assert(sa.size() == n+1);
	setText(str, n+1);
	construct(sa, str);
	if( n+1 > 0 and n+1 != size() )
		throw std::logic_error("csa_sada_prac: text size differ with sa size!");
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
uint32_t csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::get_sample_dens()const{
	if(SampleDens==0)
		return m_sample_dens;
	else
		return SampleDens;
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
void csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::set_sample_dens(const uint32_t sample_dens){
	m_sample_dens = sample_dens;
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
uint32_t csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::get_psi_sample_dens()const{
	return m_psi.get_sample_dens();
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
void csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::set_psi_sample_dens(const uint32_t sample_dens){
	m_psi.get_sample_dens();
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
void csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::setText(const unsigned char *str, size_type len){
	for(uint16_t i=0; i<256; ++i)
		m_char2comp[i] = m_comp2char[i] = m_C[i] = 0;
	m_C[256] = 0;
	if( str == NULL or len ==0 )
		return;
	const unsigned char *p = str;
	for(size_type i=0;i<len-1;++i){
//	while(*p){
		m_char2comp[*p++] = 1;
	}
	assert(m_char2comp[0] == 0);
	m_char2comp[*p] = 1;
	assert(m_char2comp[0] == 1);
	size_type value = 0;
	for(uint16_t i=0; i<256; ++i)
		if(m_char2comp[i])
			m_char2comp[i] = value++;
	m_sigma = value;
	for(uint16_t i=0; i<256; ++i)
		if(m_char2comp[i])
			m_comp2char[m_char2comp[i]] = i;
	for(uint16_t i=0; i<257; ++i) m_C[i]=0;
	for(size_type i=0; i<len; ++i) ++m_C[m_char2comp[str[i]]];
	for(uint16_t i=256;i>0; --i) m_C[i] = m_C[i-1];
	m_C[0] = 0;
	for(uint16_t i=1; i<257; ++i) m_C[i] += m_C[i-1];	
	if(m_C[256]!=len){
		std::cerr<<"m_C[256]="<<m_C[256]<<" "<<len<<std::endl;
	}
	assert(m_C[256]==len);
	assert(m_C[m_sigma+1]==len);
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
template<typename RandomAccessContainer>
void csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::construct_samples(const RandomAccessContainer &sa, const unsigned char *str){
	m_sa_sample.setIntWidth( bit_magic::l1BP(sa.size())+1 );
	m_sa_sample.resize( (sa.size()+get_sample_dens()-1)/get_sample_dens() );
	typename RandomAccessContainer::size_type i=0, idx=0;
	for(typename RandomAccessContainer::const_iterator it = sa.begin(); i < sa.size(); it += (difference_type)get_sample_dens(), i += get_sample_dens(), ++idx){
		m_sa_sample[idx] = *it;
	}
//	const uint32_t SampleDens, uint32_t InvSampleDensISA = get_sample_dens()*16;
	m_isa_sample.setIntWidth( bit_magic::l1BP(sa.size())+1 );
	m_isa_sample.resize( (sa.size()+(InvSampleDens)-1)/(InvSampleDens) );
	i = 0;
	for(typename RandomAccessContainer::const_iterator it = sa.begin(), end = sa.end(); it != end; ++it, ++i){
		if( (*it % InvSampleDens) == 0 ){
			m_isa_sample[*it/InvSampleDens] = i;
		}
	}
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
template<typename RandomAccessContainer>
void csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::construct(const RandomAccessContainer &sa, const unsigned char *str){
	construct_samples(sa, str);
	const typename RandomAccessContainer::size_type n = sa.size();
	typename RandomAccessContainer::size_type i=0;
	int_vector<> psi(n, 0, bit_magic::l1BP(n)+1 );
	if(str == NULL){
		algorithm::sa2psi(sa, psi);
	}else{
//		std::cerr<<"n="<<sa.size()<<std::endl;
//		std::cerr<<"\""<<str<<"\""<<std::endl;
		unsigned char *bwt = new unsigned char[n+1];
		i = 0;
		assert(str[n-1]=='\0');
//		bwt[i] = str[n-1];
//		++i;
		for(typename RandomAccessContainer::const_iterator it = sa.begin(), end = sa.end(); it != end; ++it, ++i){
			bwt[i] = str[(*it+n-1)%(n)];
		}
		algorithm::bwt2psi(bwt, n, psi);
		delete [] bwt;
	}
	if( psi.size() > 0 ){
		m_psi = EncVector(psi);
	}
	else{
		m_psi = EncVector();
	}
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
template<typename RandomAccessContainer>
void csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::construct(RandomAccessContainer &sa, const unsigned char *str){
	construct_samples(sa, str);
	const typename RandomAccessContainer::size_type n = sa.size();
	typename RandomAccessContainer::size_type i=0;
	if(str == NULL){
		int_vector<> psi(n, 0, bit_magic::l1BP(n)+1 );
		algorithm::sa2psi(sa, psi);
		m_psi = EncVector(psi);
	}else{
//		std::cerr<<"n="<<sa.size()<<std::endl;
//		std::cerr<<"\""<<str<<"\""<<std::endl;
//		for(size_type j=0;j<sa.size();++j){ std::cerr<<" "<<sa[j]; }
//		std::cerr<<std::endl;
		unsigned char *bwt = new unsigned char[n+1];
		bwt[n] = 0;
		i = 0;
//		std::cerr<<"n="<<n<<" strlen(str)="<<strlen((const char*)str)<<std::endl;
		assert(str[n-1]=='\0');
//		bwt[i] = str[n-1];
//		++i;
		for(typename RandomAccessContainer::const_iterator it = sa.begin(), end = sa.end(); it != end; ++it, ++i){
			bwt[i] = str[(*it+n-1)%(n)];
//			std::cerr<<" "<<*it<<" "<<(*it+n-1)%(n)<<" "<<bwt[i]<<std::endl;
		}
//		std::cerr<<"\"";
//		for(size_type j=0;j<n;++j){
//			if(bwt[j]!='\0')
//				std::cerr<<bwt[j];	
//			else
//				std::cerr<<'.';
//		}
//		std::cerr<<"\""<<std::endl;
		{// clear sa
			RandomAccessContainer tmp;
			tmp.swap(sa);
		}
		int_vector<> psi(n, 0, bit_magic::l1BP(n)+1 );
		algorithm::bwt2psi(bwt, n, psi);	
//		for(size_type j=0;j<psi.size();++j){ std::cerr<<" "<<psi[j]; }
		delete [] bwt;
		m_psi = EncVector(psi);
	}
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
typename csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::const_iterator csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::begin()const{
	return const_iterator(this, 0);
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
typename csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::const_iterator csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::end()const{
	return const_iterator(this, size());
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
inline typename csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::value_type csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::operator[](size_type i)const{
	size_type off = 0;
	while( i % get_sample_dens() ){// while i mod SampleDens != 0 (SA[i] is not sampled)
		i = m_psi[i];       // go to the position where SA[i]+1 is located
		++off;              // add 1 to the offset
	}
	value_type result = m_sa_sample[i/get_sample_dens()];
	if( result < off ){
		return m_psi.size()-(off-result);
	}
	else
		return result-off;
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
inline typename csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::value_type csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::operator()(size_type i)const{
//	const uint32_t SampleDens, uint32_t InvSampleDensISA = get_sample_dens()*16;
	value_type result = m_isa_sample[i/InvSampleDens]; // get the rightmost sampled isa value
	i = i % InvSampleDens;
	while( i-- ){
		result = m_psi[result];
	}
//	assert(((*this)[result])==j);
	return result;
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
csa_sada_prac<EncVector,SampleDens, InvSampleDens, fixedIntWidth>& csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::operator=(const csa_sada_prac<EncVector,SampleDens, InvSampleDens, fixedIntWidth> &csa){
	if(this != &csa){
		m_psi = csa.m_psi;
		m_sa_sample = csa.m_sa_sample;
		m_isa_sample = csa.m_isa_sample;
		for(int i=0; i<256; ++i){
		   	m_char2comp[i] = csa.m_char2comp[i];
			m_comp2char[i] = csa.m_comp2char[i];
			m_C[i]		 = csa.m_C[i];
			m_sigma		 = csa.m_sigma;
		}
		m_C[256] = csa.m_C[256];
	}
	return *this;
}


template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
typename csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::size_type csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::serialize(std::ostream &out)const{
	size_type written_bytes = 0;
	written_bytes += m_psi.serialize(out);
	written_bytes += m_sa_sample.serialize(out);
	written_bytes += m_isa_sample.serialize(out);
	size_type wb   = sizeof(m_char2comp[0])*256;
	out.write((char*)m_char2comp, wb);
	written_bytes += wb;
	wb			   = sizeof(m_comp2char[0])*256;
	out.write((char*)m_comp2char, wb);
	written_bytes += wb;
	wb			   = sizeof(m_C[0])*257;
	out.write((char*)C, wb);
	written_bytes += wb;
	out.write((char*)&m_sigma, sizeof(m_sigma));
	wb += sizeof(m_sigma);
	return written_bytes;
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
void csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::load(std::istream &in){
	m_psi.load(in);
	m_sa_sample.load(in);
	m_isa_sample.load(in);
	in.read((char*)m_char2comp, sizeof(m_char2comp[0])*256);
	in.read((char*)m_comp2char, sizeof(m_comp2char[0])*256);
	in.read((char*)m_C, sizeof(m_C[0])*257);
	in.read((char*)&m_sigma, sizeof(m_sigma));
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
void csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::swap(csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth> &csa){
	if(this != &csa){
		m_psi.swap(csa.m_psi);
		m_sa_sample.swap(csa.m_sa_sample);
		m_isa_sample.swap(csa.m_isa_sample);
		for(uint16_t i=0;i<256;++i){
			std::swap(m_char2comp[i], csa.m_char2comp[i]);
			std::swap(m_comp2char[i], csa.m_comp2char[i]);
			std::swap(m_C[i], csa.m_C[i]);
		}
		std::swap(m_C[256], csa.m_C[256]);
		std::swap(m_sigma, csa.m_sigma);
	}
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
bool csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::operator==(const csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth> &csa)const{
	for(uint16_t i=0;i<256;++i)
		if(m_char2comp[i] != csa.m_char2comp[i] or m_comp2char[i] != csa.m_comp2char[i] or m_C[i] != csa.m_C[i])
			return false;
	return m_psi == csa.m_psi and m_sa_sample == csa.m_sa_sample and m_isa_sample == csa.m_isa_sample and m_C[256] == csa.m_C[256] and m_sigma == csa.m_sigma;
}

template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
bool csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::operator!=(const csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth> &csa)const{
	return !(*this == csa); 
//	return m_psi != csa.m_psi or m_sa_sample != csa.m_sa_sample or m_isa_sample != csa.m_isa_sample;
}

} // end namespace sdsl

#endif
