/* sdsl - succinct data structures library
    Copyright (C) 2009 Simon Gog 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file csa_wt.hpp
    \brief csa_wt.hpp contains an implementation of the compressed suffix array based on a wavelet tree. 
	\author Simon Gog
*/
#ifndef INCLUDED_SDSL_CSA_WT
#define INCLUDED_SDSL_CSA_WT

#include "suffixarrays.hpp"
#include "wavelet_tree.hpp"
#include "algorithms.hpp"
#include "iterators.hpp"
#include "util.hpp"
#include "lcp.hpp"
#include <iostream>
#include <algorithm> // for std::swap
#include <cassert>
#include <cstring> // for strlen
#include <iomanip>
#include <iterator>

namespace sdsl{



template<uint8_t fixedIntWidth>
struct csa_wt_trait{
	typedef int_vector<0> int_vector_type;
};

template<>
struct csa_wt_trait<32>{
	typedef int_vector<32> int_vector_type;
};

template<>
struct csa_wt_trait<64>{
	typedef int_vector<64> int_vector_type;
};

//! A wrapper class for the \f$\Psi\f$ function for (compressed) suffix arrays that are based on a wavelet tree (like sdsl::csa_wt).
template<class CsaWT>
class psi_of_csa_wt{
	public:
	typedef typename CsaWT::value_type value_type;	
	typedef typename CsaWT::size_type size_type;	
	private:
		CsaWT *m_csa_wt; //<- pointer to the (compressed suffix array that is based on a wavelet tree) 
	public:

	//! Constructor	
	psi_of_csa_wt(CsaWT *csa_wt=NULL){
		m_csa_wt = csa_wt;
	}

	//! Copy constructor
	psi_of_csa_wt(const psi_of_csa_wt &csa){
		m_csa_wt = csa.m_csa_wt; 
	}

	//! Calculate the \f$\Psi\f$ value at position i.
	/*!	\param i The index for which the \f$\Psi\f$ value should be calculated, \f$i\in [0..size()-1]\f$. 
	 *	\par Time complexity
	 *		\f$ \Order{\log |\Sigma|} \f$ 
	 */
	value_type operator[](size_type i)const{
		assert(m_csa_wt!=NULL);
		assert(i<size());
		size_type idx=1;
		if( m_csa_wt->sigma < 16 ){ //<- if sigma is small search linear
			while( m_csa_wt->C[idx] <= i ){
				++idx;
			}
		}else{ //<- binary search on C takes \f$ \log |\Sigma| \f$ time
			size_type begin=1, end=m_csa_wt->sigma+1; // begin inclusive, end exclusive
			while(end>begin){
				idx = (begin+end)/2;
				if( m_csa_wt->C[idx] <= i ){
					begin = idx+1;
				}else{
					end = idx;
				}
			}
			idx = end;
		}
		// select takes also \f$ \log |\Sigma| \f$ time
		return m_csa_wt->m_wavelet_tree.select( i-m_csa_wt->C[idx-1]+1, m_csa_wt->comp2char[idx-1]);
	}

	//! Calculate the LF mapping at position i.
	/*! The LF mapping function is the inverse to the \f$\Psi\f$ function. That is \f$LF[\Psi(i)]=i\f$. 
	 *  \param i The index for which the LF value should be calculated, \f$i\in [0..size()-1]\f$.
	 */ 
	value_type operator()(size_type i)const{
		assert(m_csa_wt != NULL);
		assert(i<size());
		// (1) get the ith character in the bwt, this takes \f$\log |\Sigma| \f$ time with use of the wavelet tree
		value_type c = m_csa_wt->m_wavelet_tree[i]; 
		// (2) next count how many symbos c are in the prefix [0..i-1] of the bwt, this takes \f$\log |\Sigma| \f$ time with use of the wavelet tree.
		size_type j = m_csa_wt->m_wavelet_tree.rank(i, c);
		// note that (1) and (2) can be calculated in only one request to the wavelet tree, maybe TODO
		// (3) calculate the position of the j+1 th c in the sorted string of the bwt in constant time.
		return m_csa_wt->C[ m_csa_wt->char2comp[c] ] + j;
	}

	//! Assignment operator
	psi_of_csa_wt& operator=(const psi_of_csa_wt& psi){
		if( this != &psi ){
			m_csa_wt = psi.m_csa_wt;
		}
		return *this;
	}

	//! Equality operator
	/*! return Always true, since all wrapper objects are equal only the reference to the supported csa differs.
	 */ 
	bool operator==(const psi_of_csa_wt& psi){
		return true;
	}

	//! Returns the size of the \f$\Psi\f$ function.
	size_type size()const{
		assert(m_csa_wt!=NULL);
		return m_csa_wt->size();
	}

	//! Swap operation require by the STL.
	void swap(psi_of_csa_wt &psi){
		if( this != &psi){
			;// std::swap(m_csa_wt, psi.m_csa_wt); // do not exchange the supported structure!
		}
	}

	// Get the number of sampled \f$\Psi\f$ values. In the wavelet tree approach the number of sampled (i.e. explicitly stored) \f$\Psi\f$ values is zero.
	uint32_t get_sample_dens()const{
		return 1;
	}
};

//! A class for the Compressed Suffix Array (CSA) based on a Wavelet Tree (WT) for the Burrow Wheeler Transform of the orignal text. 
/*! The CSA is parameterized with an WavletTree, the sample density SampleDens (\f$s_{SA}\f$), and the sample density for inverse suffix array entries (\f$s_{SA^{-1}}\f$).
  * I.e. every \f$s_{SA}th\f$ value from the original suffix array is explicitely stored with \f$\log n\f$ bits.
  * 
  * \todo example, code example
  *  \sa csa_sada_prac
  * @ingroup csa
 */
template<class WaveletTree = wavelet_tree<>, uint32_t SampleDens = 12, uint32_t InvSampleDens = 12,  uint8_t fixedIntWidth = 32>
class csa_wt{
	public:
	typedef uint64_t											 value_type;	// STL Container requirement
	typedef random_access_const_iterator<csa_wt>		 		 const_iterator;// STL Container requirement
	typedef const_iterator 										 iterator;		// STL Container requirement
	typedef const value_type									 const_reference;
	typedef const_reference										 reference;
	typedef const_reference*									 pointer;
	typedef const pointer										 const_pointer;
	typedef size_t 												 size_type;		// STL Container requirement
	typedef ptrdiff_t  											 difference_type; // STL Container requirement
	typedef psi_of_csa_wt<csa_wt>								 enc_vector_type;
	typedef WaveletTree											 wavelet_tree_type;
	typedef const unsigned char*								 pattern_type;

	friend class psi_of_csa_wt<csa_wt>;
//	static const uint32_t sample_dens = SampleDens;
	private:
		WaveletTree		m_wavelet_tree; // the wavelet tree
		enc_vector_type m_psi;  // psi function
		typename csa_wt_trait<fixedIntWidth>::int_vector_type m_sa_sample; // suffix array samples
		typename csa_wt_trait<fixedIntWidth>::int_vector_type m_isa_sample; // inverse suffix array samples
		unsigned char	m_char2comp[256];
		unsigned char 	m_comp2char[256];
		size_type       m_C[257];
		uint8_t			m_sigma;
		uint32_t m_sample_dens; // additional to SampleDens value

		template<typename RandomAccessContainer>
		void construct(const RandomAccessContainer &sa, const unsigned char *str);

		template<typename RandomAccessContainer>
		void construct_samples(const RandomAccessContainer &sa, const unsigned char *str);

		void setText(const unsigned char *str, size_type len);
	public:
		const unsigned char *char2comp;
		const unsigned char *comp2char;
		const size_type *C;
		const uint8_t &sigma;
		const enc_vector_type &psi;
		const wavelet_tree_type &wavelet_tree;
		//! Default Constructor
		csa_wt():char2comp(m_char2comp), comp2char(m_comp2char), C(m_C), sigma(m_sigma) ,psi(m_psi), wavelet_tree(m_wavelet_tree){}
		//! Default Destructor
		~csa_wt(){}
		//! Copy constructor
		csa_wt(const csa_wt &csa): m_sa_sample(csa.m_sa_sample), m_isa_sample(csa.m_isa_sample),  char2comp(m_char2comp), comp2char(m_comp2char), C(m_C), sigma(m_sigma), psi(m_psi), wavelet_tree(m_wavelet_tree)
		{ 
			m_wavelet_tree			= csa.m_wavelet_tree;
			m_sigma			= csa.m_sigma;
			for(int i=0; i<256; ++i){
				m_char2comp[i] 	= csa.m_char2comp[i];
				m_comp2char[i]	= csa.m_comp2char[i];
				m_C[i]			= csa.m_C[i];
			}
			m_C[256] = csa.m_C[256];
			m_psi = enc_vector_type(this);
		}

		//! Construct the csa_wt from another compressed or uncompressed suffix array
		template<typename RandomAccessContainer>
		csa_wt(const RandomAccessContainer &sa, const unsigned char *str);

		//! Constructor for the CSA taking a string for that the CSA should be calculated
		template<class Lcp_type>
		csa_wt(const unsigned char *str, Lcp_type *lcp);

		//! Number of elements in the \f$\CSA\f$.
		/*! Required for the Container Concept of the STL.
		 *  \sa max_size, empty
		 *  \par Time complexity
		 *      \f$ \Order{1} \f$
		 */
		size_type size()const{ return m_wavelet_tree.size();}

		//! Returns the largest size that csa_wt can ever have.
		/*! Required for the Container Concept of the STL.
		 *  \sa size
		 */
		static size_type max_size(){ return bit_vector::max_size(); }

		//! Returns if the data strucutre is empty.
		/*! Required for the Container Concept of the STL.
		 * \sa size
		 */
		bool empty()const{ return m_wavelet_tree.empty();}

		//! Swap method for csa_wt
		/*! The swap method can be defined in terms of assignment.
			This requires three assignments, each of which, for a container type, is linear
			in the container's size. In a sense, then, a.swap(b) is redundant.
			This implementation guaranties a run-time complexity that is constant rather than linear.
			\param csa csa_wt to swap.

			Required for the Assignable Conecpt of the STL.
		  */
		void swap(csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth> &csa);

		//! Returns a const_iterator to the first element.
		/*! Required for the STL Container Concept.
		 *  \sa end
		 */
		const_iterator begin()const;

		//! Returns a const_iterator to the element after the last element.
		/*! Required for the STL Container Concept.
		 *  \sa begin.
		 */
		const_iterator end()const;

		//! []-operator
		/*! \param i Index of the value. \f$ i \in [0..size()-1]\f$.
		 * Required for the STL Random Access Container Concept.
		 * \par Time complexity
		 *      \f$ \Order{s_{SA}\cdot t_{\Psi}} \f$, where every \f$s_{SA}\f$th suffix array entry is sampled and \f$t_{\Psi}\f$ 
		 *           is the access time for an element in the \f$\Psi\f$-function.
		 */
		inline value_type operator[](size_type i)const;

		//! ()-operator return inverse suffix array values
		/*! \param i Index of the value. \f$ i \in [0..size()-1]\f$.
		 *   \par Time complexity
		 *      \f$ \Order{s_{SA^{-1}}\cdot t_{\Psi}} \f$, where every \f$s_{SA^{-1}}\f$th suffix array entry is sampled and \f$t_{\Psi}\f$ 
		 *           is the access time for an element in the \f$\Psi\f$-function.
		 */
		inline value_type operator()(size_type i)const;

		//! Assignment Operator.
		/*!
		 *	Required for the Assignable Concept of the STL.
		 */
		csa_wt& operator=(const csa_wt &csa);

		//! Equality Operator
		/*! Two Instances of csa_wt are equal if
		 *  all their members are equal.
		 *  \par Required for the Equality Comparable Concept of the STL.
		 *  \sa operator!=
		 */
		bool operator==(const csa_wt &csa)const;

		//! Unequality Operator
		/*! Two Instances of csa_wt are equal if
		 *  not all their members are equal.
		 *  \par Required for the Equality Comparable Concept of the STL.
		 *  \sa operator==
		 */
		bool operator!=(const csa_wt &csa)const;

		//! Serialize to a stream.
		/*! \param out Outstream to write the data structure.
		 *  \return The number of written bytes.
		 */
		size_type serialize(std::ostream &out) const;

		//! Load from a stream.
		/*! \param in Inputstream to load the data structure from.
		 */
		void load(std::istream &in);

		uint32_t get_sample_dens() const;
		void set_sample_dens(const uint32_t sample_dens);

		uint32_t get_psi_sample_dens() const;
		void set_psi_sample_dens(const uint32_t sample_dens);
};

// == template functions ==

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
template<class Lcp_type>
csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::csa_wt(const unsigned char *str, Lcp_type *lcp):char2comp(m_char2comp), comp2char(m_comp2char), C(m_C), sigma(m_sigma), psi(m_psi), wavelet_tree(m_wavelet_tree){
	size_type n = strlen((const char*)str);
	
/*
	int_vector<> sa(n+1, 0, bit_magic::l1BP(n+1)+1);
	algorithm::calculate_sa(str, n+1, sa);	 // calculate the suffix array sa of str
	assert(sa.size() == n+1);
	setText(str, n+1);
	construct(sa, str);
	for(int i=0; i<26; ++i)
	{
		std::cerr<<m_psi[i]<<" ";
	}
	std::cerr<<std::endl;

	for(int i=0; i<26; ++i)
	{
		std::cerr<<sa[i]<<" ";
	}
	std::cerr<<std::endl;
	if( n+1 > 0 and n+1 != size() )
		throw std::logic_error(util::demangle(typeid(this).name())+": text size differ with sa size!");
*/
	
	setText(str, n+1);
	int *x = new int[n+1];
	int *p = new int[n+1];
	for(int i=0; i<n+1; ++i)
	{
		x[i] = char2comp[str[i]];
	}
	std::cerr<<"# begin constructing sa"<<std::endl;
	std::cerr<<"# sigma="<<(int)sigma<<std::endl;
	std::cerr<<"# sample_dens="<<get_sample_dens()<<std::endl;
	suffixsort(x, p, n, sigma, 1);
//	for(int i=0; i<26; ++i)
//	{
//		std::cerr<<x[(p[i]+1)%(n+1)]<<" ";
//	}
//	std::cerr<<std::endl;
//	for(int i=0; i<26; ++i)
//	{
//		std::cerr<<p[i]<<" ";
//	}
//	std::cerr<<std::endl;
	std::cerr<<"# end of construction sa"<<std::endl;

	m_sa_sample.setIntWidth( bit_magic::l1BP(n+1)+1 );
	m_sa_sample.resize( (n+1+get_sample_dens()-1)/get_sample_dens() );
	m_isa_sample.setIntWidth( bit_magic::l1BP(n+1)+1 );
	m_isa_sample.resize( (n+1+(InvSampleDens)-1)/(InvSampleDens) );
	for(int i = 0, idx=0; i < n+1; i += InvSampleDens, ++idx){
		m_isa_sample[idx] = x[i];
	}
	for(int i = 0, idx=0; i < n+1; i += get_sample_dens(), ++idx){
		m_sa_sample[idx] = p[i];
	}

	std::cerr<<"# samples written"<<std::endl;

//	const typename RandomAccessContainer::size_type n = sa.size();
	unsigned char *bwt = new unsigned char[n+2];
	assert(str[n]=='\0');
	for(size_type i = 0; i <= n; ++i){
		bwt[i] = str[(p[i]+n)%(n+1)];
	}
	std::cerr<<"# bwt calculated"<<std::endl;
	bwt[n+1] = '\0';
	m_wavelet_tree = WaveletTree(bwt, n+1);
	std::cerr<<"# wavelet tree calculated"<<std::endl;
	m_psi = enc_vector_type(this);
#ifdef SDSL_DEBUG
	std::cerr<<"# check suffix array and psi"<<std::endl;
	for(int i=0; i<size(); ++i)
	{
		if( p[i] != (*this)[i] ){
			std::cerr<<"i="<<i<<" p[i]="<<p[i]<<" != (*this)[i]= "<<(*this)[i]<<std::endl;
			assert(false);
		}
		if( x[(p[i]+1)%size()] != this->psi[i]){
			std::cerr<<"i="<<i<<"x( p[i]+1 mod size())="<<x[(p[i]+1)%size()]<<" != this->psi[i]= "<<this->psi[i]<<std::endl;
			assert(false);
		}
		if( x[(p[i]+size()-1)%size()] != this->psi(i)){
			std::cerr<<"i="<<i<<"x( p[i]-1 mod size())="<<x[(p[i]+size()-1)%size()]<<" != this->psi(i)= "<<this->psi(i)<<std::endl;
			assert(false);
		}
		if(i%1000000==0){
			std::cerr<<".";
		}
	}
	std::cerr<<std::endl;
	std::cerr<<"sa checked"<<std::endl;
#endif	
	csa_wrapper_for_sa csa_wrapper(p,x,size());
	lcp_trait<csa_wt, Lcp_type>::construct_lcp(*lcp, *this, csa_wrapper, str);
	delete [] bwt;
	delete [] x;	
	delete [] p;
	std::cerr<<"# leave"<<std::endl;
	
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
template<typename RandomAccessContainer>
csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::csa_wt(const RandomAccessContainer &sa, const unsigned char *str):char2comp(m_char2comp), comp2char(m_comp2char),C(m_C), sigma(m_sigma), psi(m_psi), wavelet_tree(m_wavelet_tree){
	size_type n = 1;
	if(str != NULL){
		n = strlen((const char*)str);
	}
	setText(str, n+1);
	assert(sa.size() == n+1);
	construct(sa, str);
	if( n+1 > 0 and n+1 != size() )
		throw std::logic_error(util::demangle(typeid(this).name())+": text size differ with sa size! ");
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
uint32_t csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::get_sample_dens()const{
	if(SampleDens==0)
		return m_sample_dens;
	else
		return SampleDens;
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
void csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::set_sample_dens(const uint32_t sample_dens){
	m_sample_dens = sample_dens;
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
uint32_t csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::get_psi_sample_dens()const{
	return m_psi.get_sample_dens();
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
void csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::set_psi_sample_dens(const uint32_t sample_dens){
	m_psi.get_sample_dens();
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
void csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::setText(const unsigned char *str, size_type len){
	for(uint16_t i=0; i<256; ++i)
		m_char2comp[i] = m_comp2char[i] = m_C[i] = 0;
	m_C[256] = 0;
	if( str == NULL or len ==0 )
		return;
	const unsigned char *p = str;
	for(size_type i=0;i<len-1;++i){
		m_char2comp[*p++] = 1;
	}
	assert(m_char2comp[0] == 0);
	m_char2comp[*p] = 1;
	assert(m_char2comp[0] == 1);
	size_type value = 0;
	for(uint16_t i=0; i<256; ++i)
		if(m_char2comp[i])
			m_char2comp[i] = value++;
	m_sigma = value;
	for(uint16_t i=0; i<256; ++i)
		if(m_char2comp[i])
			m_comp2char[m_char2comp[i]] = i;
	for(uint16_t i=0; i<257; ++i) m_C[i]=0;
	for(size_type i=0; i<len; ++i) ++m_C[m_char2comp[str[i]]];
	for(uint16_t i=256;i>0; --i) m_C[i] = m_C[i-1];
	m_C[0] = 0;
	for(uint16_t i=1; i<257; ++i) m_C[i] += m_C[i-1];	
	if(m_C[256]!=len){
		std::cerr<<"m_C[256]="<<m_C[256]<<" "<<len<<std::endl;
	}
	assert(m_C[256]==len);
	assert(m_C[m_sigma+1]==len);
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
template<typename RandomAccessContainer>
void csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::construct_samples(const RandomAccessContainer &sa, const unsigned char *str){
	m_sa_sample.setIntWidth( bit_magic::l1BP(sa.size())+1 );
	m_sa_sample.resize( (sa.size()+get_sample_dens()-1)/get_sample_dens() );
	typename RandomAccessContainer::size_type i=0, idx=0;
	for(typename RandomAccessContainer::const_iterator it = sa.begin(); i < sa.size(); it += (difference_type)get_sample_dens(), i += get_sample_dens(), ++idx){
		m_sa_sample[idx] = *it;
	}
	m_isa_sample.setIntWidth( bit_magic::l1BP(sa.size())+1 );
	m_isa_sample.resize( (sa.size()+(InvSampleDens)-1)/(InvSampleDens) );
	i = 0;
	for(typename RandomAccessContainer::const_iterator it = sa.begin(), end = sa.end(); it != end; ++it, ++i){
		if( (*it % InvSampleDens) == 0 ){
			m_isa_sample[*it/InvSampleDens] = i;
		}
	}
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
template<typename RandomAccessContainer>
void csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::construct(const RandomAccessContainer &sa, const unsigned char *str){
	construct_samples(sa, str);
	const typename RandomAccessContainer::size_type n = sa.size();
	typename RandomAccessContainer::size_type i=0;
	if(str == NULL){
		throw std::logic_error(util::demangle(typeid(this).name())+": text is needed to construct the csa_wt!");
	}else{
		unsigned char *bwt = new unsigned char[n+1];
		i = 0;
		assert(str[n-1]=='\0');
		for(size_type i = 0; i < sa.size(); ++i){
			bwt[i] = str[(sa[i]+n-1)%(n)];
		}
		bwt[n] = '\0';
		m_wavelet_tree = WaveletTree(bwt, n);
		m_psi = enc_vector_type(this);
		delete [] bwt;
	}
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
typename csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::const_iterator csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::begin()const{
	return const_iterator(this, 0);
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
typename csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::const_iterator csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::end()const{
	return const_iterator(this, size());
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
inline typename csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::value_type csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::operator[](size_type i)const{
	size_type off = 0;
	while( i % get_sample_dens() ){// while i mod SampleDens != 0 (SA[i] is not sampled)
		i = m_psi[i];       // go to the position where SA[i]+1 is located
		++off;              // add 1 to the offset
	}
	assert(get_sample_dens()!=0);
//	std::cerr<<" i="<<i<<" get_sample_dens="<<get_sample_dens()<<" m_sa_sample.size()="<<m_sa_sample.size()<<std::endl;
	assert(i/get_sample_dens() < m_sa_sample.size());
	value_type result = m_sa_sample[i/get_sample_dens()];
//	std::cerr<<"result="<<result<<std::endl;
//	std::cerr<<"m_psi.size()="<<m_psi.size()<<" size()="<<size()<<std::endl;
	if( result < off ){
		return size()-(off-result);
	}
	else
		return result-off;
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
inline typename csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::value_type csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::operator()(size_type i)const{
//	const uint32_t SampleDens, uint32_t InvSampleDensISA = get_sample_dens()*16;
	value_type result = m_isa_sample[i/InvSampleDens]; // get the rightmost sampled isa value
	i = i % InvSampleDens;
	while( i-- ){
		result = m_psi[result];
	}
//	assert(((*this)[result])==j);
	return result;
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
csa_wt<WaveletTree,SampleDens, InvSampleDens, fixedIntWidth>& csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::operator=(const csa_wt<WaveletTree,SampleDens, InvSampleDens, fixedIntWidth> &csa){
	if(this != &csa){
		m_wavelet_tree = csa.m_wavelet_tree;
		m_sa_sample = csa.m_sa_sample;
		m_isa_sample = csa.m_isa_sample;
		for(int i=0; i<256; ++i){
		   	m_char2comp[i] = csa.m_char2comp[i];
			m_comp2char[i] = csa.m_comp2char[i];
			m_C[i]		 = csa.m_C[i];
			m_sigma		 = csa.m_sigma;
		}
		m_C[256] = csa.m_C[256];
		m_psi = enc_vector_type(this);
	}
	return *this;
}


template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
typename csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::size_type csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::serialize(std::ostream &out)const{
	size_type written_bytes = 0;
	written_bytes += m_wavelet_tree.serialize(out);
	written_bytes += m_sa_sample.serialize(out);
	written_bytes += m_isa_sample.serialize(out);
	size_type wb   = sizeof(m_char2comp[0])*256;
	out.write((char*)m_char2comp, wb);
	written_bytes += wb;
	wb			   = sizeof(m_comp2char[0])*256;
	out.write((char*)m_comp2char, wb);
	written_bytes += wb;
	wb			   = sizeof(m_C[0])*257;
	out.write((char*)C, wb);
	written_bytes += wb;
	out.write((char*)&m_sigma, sizeof(m_sigma));
	wb += sizeof(m_sigma);
	return written_bytes;
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
void csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::load(std::istream &in){
	m_wavelet_tree.load(in);
	m_sa_sample.load(in);
	m_isa_sample.load(in);
	in.read((char*)m_char2comp, sizeof(m_char2comp[0])*256);
	in.read((char*)m_comp2char, sizeof(m_comp2char[0])*256);
	in.read((char*)m_C, sizeof(m_C[0])*257);
	in.read((char*)&m_sigma, sizeof(m_sigma));
	m_psi = enc_vector_type(this);
}
 
template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
void csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::swap(csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth> &csa){
	if(this != &csa){
		m_wavelet_tree.swap(csa.m_wavelet_tree);
		m_sa_sample.swap(csa.m_sa_sample);
		m_isa_sample.swap(csa.m_isa_sample);
		for(uint16_t i=0;i<256;++i){
			std::swap(m_char2comp[i], csa.m_char2comp[i]);
			std::swap(m_comp2char[i], csa.m_comp2char[i]);
			std::swap(m_C[i], csa.m_C[i]);
		}
		std::swap(m_C[256], csa.m_C[256]);
		std::swap(m_sigma, csa.m_sigma);
		m_psi.swap(csa.m_psi);
	}
}

/* TODO
template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
bool csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::operator==(const csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth> &csa)const{
	for(uint16_t i=0;i<256;++i)
		if(m_char2comp[i] != csa.m_char2comp[i] or m_comp2char[i] != csa.m_comp2char[i] or m_C[i] != csa.m_C[i])
			return false;
	return m_psi == csa.m_psi and m_sa_sample == csa.m_sa_sample and m_isa_sample == csa.m_isa_sample and m_C[256] == csa.m_C[256] and m_sigma == csa.m_sigma;
}

template<class WaveletTree, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
bool csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth>::operator!=(const csa_wt<WaveletTree, SampleDens, InvSampleDens, fixedIntWidth> &csa)const{
	return !(*this == csa); 
//	return m_psi != csa.m_psi or m_sa_sample != csa.m_sa_sample or m_isa_sample != csa.m_isa_sample;
}
*/

} // end namespace sdsl

#endif
