/* sdsl - succinct data structures library
    Copyright (C) 2008 Simon Gog 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file suffixarrays.hpp
    \brief suffixarrays.hpp contains generic classes for different suffix array classes.
	\author Simon Gog
*/
#ifndef INCLUDED_SDSL_SUFFIXARRAYS
#define INCLUDED_SDSL_SUFFIXARRAYS

#include "int_vector.hpp" 

extern "C"{
#include "qsufsort.h"
}

//! A wrapper class for the \f$\Psi\f$ function for (compressed) suffix arrays which provide also the inverse suffix array values (like sdsl::csa_uncompressed).	
template<class Csa>
class psi_of_csa{
	public:
	typedef typename Csa::value_type value_type;
	typedef typename Csa::size_type size_type;
	private:
		Csa *m_csa; //<- pointer to the (full text index) suffix array 
	public:
	
	//! Constructor
	psi_of_csa(Csa *csa=NULL){
		m_csa = csa;
	}	

	// Copy constructor
	psi_of_csa(const psi_of_csa &csa){
		m_csa = csa.m_csa;
	}

	//! Calculate the \f$\Psi\f$ value at position i.
	/*!	\param i The index for which the \f$\Psi\f$ value should be calculated, \f$i\in [0..size()-1]\f$. 
	 *	\par Time complexity
	 *		\f$ \Order{\saaccess+\isaaccess} \f$ 
	*/
	value_type operator[](size_type i)const{
		assert(m_csa!=NULL);
		assert(i<size());
		// \f$\Psi[i] = SA^{-1}[SA[i]+1 \mod n]\f$, where \f$n\f$ is the length of the suffix array SA
		return (*m_csa)( ((*m_csa)[i]+1) % m_csa->size()  );
	} 

	//! Calculate the LF mapping at position i.
	/*! The LF mapping function is the inverse to the \f$\Psi\f$ function. That is \f$LF[\Psi(i)]=i\f$. 
	 *  \param i The index for which the LF value should be calculated, \f$i\in [0..size()-1]\f$.
	 */ 
	value_type operator()(size_type i)const{
		assert(m_csa!=NULL);
		assert(i<size());
		// \f$LF[i] = SA^{-1}[SA[i]-1 \mod n]\f$, where \f$n\f$ is the length of the suffix array SA
		return (*m_csa)( ((*m_csa)[i]+m_csa->size()-1) % m_csa->size() );	
	}

	//! Assignment operator
	psi_of_csa& operator=(const psi_of_csa& psi){
		if( this != &psi ){
			m_csa = psi.m_csa;
		}
		return *this;
	}

	//! Equality operator
	/*! return Always true, since all wrapper objects are equal only the reference to the supported csa differs.
	 */ 
	bool operator==(const psi_of_csa& psi)const{
		return true;
	}

	//! Returns the size of the \f$\Psi\f$ function.
	size_type size()const{
		return m_csa->size();
	}

	//! Swap operation require by the STL.
	void swap(psi_of_csa &psi){
		if( this != &psi){
			;// std::swap(m_csa, psi.m_csa); // do not exchange the supported structure!
		}
	}

	// Get the number of sampled \f$\Psi\f$ values. In this wrapper the number of sampled (i.e. explicitly stored) \f$\Psi\f$ values is zero.
	uint32_t get_sample_dens()const{
		return 1;
	}
	
};

class csa_wrapper_for_sa{
	public:
	typedef uint64_t size_type;
	typedef uint64_t value_type;
	private:
	const int *m_sa, *m_isa;
	const size_type m_n;
	typedef psi_of_csa<csa_wrapper_for_sa> psi_type;
	public:
	const psi_type psi;

	csa_wrapper_for_sa(int *sa, int *isa, size_type n):m_sa(sa), m_isa(isa), m_n(n), psi(this){
	}	

	size_type size()const{
		return m_n;
	}

	size_type operator[](size_type i)const{
		assert(i<m_n);
		return m_sa[i];
	}

	size_type operator()(size_type i)const{
		assert(i<m_n);
		return m_isa[i];
	}
};

/** \defgroup csa Compressed Suffix Arrays (CSA) */


#include "csa_sada_theo.hpp"
#include "csa_sada_prac.hpp"
#include "csa_uncompressed.hpp"
#include "csa_wt.hpp"

#endif
